#define std_db_BTree_IMPORT
#define std_lang_Chunk_IMPORT
#define std_lang_ImmutableChunk_IMPORT
#define std_lang_Object_IMPORT

#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#ifdef WITH_BDB4
#include <db_185.h>
#else
#include <db.h>
#endif
#include <std/db/impl/BTree.h>
#include <std/lang/Chunk.h>
#include <std/lang/ImmutableChunk.h>

#define NOT_PRESERVED (1 << 0)
#define READONLY (1 << 1)

struct std_db_BTree_SeqOption {
    u_int flags;
};

static void
Chunk_setDBT(/*@out@*/ struct Chunk *c, DBT t)
{
    c->data = t.data;
    c->size = t.size;
}

static void
DBT_setImmutableChunk(/*@out@*/ DBT *t, struct ImmutableChunk c)
{
    t->data = (void *)c.data;
    t->size = c.size;
}

static void
freeImpl(struct impl_BTree *impl)
{
    free(impl);
}

union BTree
BTree_create(const char *file, int flags, int mode)
{
    struct impl_BTree *impl;

    if ((impl = (struct impl_BTree *)malloc(sizeof(*impl))) == NULL) {
        return (union BTree)Object_NULL;
    }
    impl->db = NULL;
    impl->flags = 0;

#ifdef WITH_BDB4
    if (file == NULL) {
	impl->flags |= NOT_PRESERVED;
    }
    if ((flags & O_ACCMODE) == O_RDONLY) {
	impl->flags |= READONLY;
    }
    if (file != NULL
	&& flags & O_CREAT
	&& (flags & O_ACCMODE) == O_RDONLY) {
	flags &= ~O_ACCMODE;
	flags |= O_RDWR;
    }
#endif
    if ((impl->db = dbopen(file, flags, mode, DB_BTREE, NULL)) == NULL) {
	freeImpl(impl);
	return (union BTree)Object_NULL;
    }
    return (union BTree)impl;
}

int
BTree_close(union BTree tree)
{
    struct impl_BTree *impl = tree.impl;
    DB *db = impl->db;
    int s;
    
    s = db->close(db);
    freeImpl(impl);
    return s;
}

int
BTree_sync(union BTree tree)
{
    DB *db = tree.impl->db;

    return db->sync(db, 0);
}

int
BTree_getFileDescriptor(union BTree tree)
{
    struct impl_BTree *impl = tree.impl;
    DB *db = impl->db;

#ifdef WITH_BDB4
    if (impl->flags & NOT_PRESERVED) {
	errno = ENOENT;
	return -1;
    }
#endif
    return db->fd(db);
}

int
BTree_del(union BTree tree, struct ImmutableChunk key)
{
    struct impl_BTree *impl = tree.impl;
    DB *db = impl->db;
    DBT k;

#ifdef WITH_BDB4
    if (impl->flags & READONLY) {
	errno = EPERM;
	return -1;
    }
#endif
    DBT_setImmutableChunk(&k, key);
    return db->del(db, &k, 0);
}

int
BTree_delCursor(union BTree tree)
{
    struct impl_BTree *impl = tree.impl;
    DB *db = impl->db;
    DBT k;

#ifdef WITH_BDB4
    if (impl->flags & READONLY) {
	errno = EPERM;
	return -1;
    }
#endif
    k.data = NULL;
    k.size = 0;
    return db->del(db, &k, R_CURSOR);
}

int
BTree_get(union BTree tree, struct ImmutableChunk key,
	  /*@out@*/ struct Chunk *val)
{
    DB *db = tree.impl->db;
    DBT k, v;
    int s;

    DBT_setImmutableChunk(&k, key);
    if ((s = db->get(db, &k, &v, 0)) != 0)
	return s;
    Chunk_setDBT(val, v);
    return 0;
}

static int
put(union BTree tree, struct ImmutableChunk key, struct ImmutableChunk val,
    u_int flags)
{
    struct impl_BTree *impl = tree.impl;
    DB *db = impl->db;
    DBT k, v;

#ifdef WITH_BDB4
    if (impl->flags & READONLY) {
	errno = EPERM;
	return -1;
    }
#endif
    DBT_setImmutableChunk(&k, key);
    DBT_setImmutableChunk(&v, val);
    return db->put(db, &k, &v, flags);
}

int
BTree_put(union BTree tree, struct ImmutableChunk key,
	  struct ImmutableChunk val)
{
    return put(tree, key, val, 0);
}

int
BTree_putNoOverwrite(union BTree tree, struct ImmutableChunk key,
		     struct ImmutableChunk val)
{
    return put(tree, key, val, R_NOOVERWRITE);
}

#if 0
int
BTree_putCursor(union BTree tree, struct ImmutableChunk key,
		struct ImmutableChunk val)
{
    return put(tree, key, val, R_CURSOR);
}

int
BTree_putSetCursor(union BTree tree, struct ImmutableChunk key,
		   struct ImmutableChunk val)
{
    return put(tree, key, val, R_SETCURSOR);
}
#endif

int
BTree_seqCursor(union BTree tree, struct ImmutableChunk partialkey,
		/*@out@*/ struct Chunk *key, /*@out@*/ struct Chunk *val)
{
    DB *db = tree.impl->db;
    DBT k, v;
    int s;

    DBT_setImmutableChunk(&k, partialkey);
    if ((s = db->seq(db, &k, &v, R_CURSOR)) != 0)
	return s;
    Chunk_setDBT(key, k);
    Chunk_setDBT(val, v);
    return 0;
}

int
BTree_seq(union BTree tree,
	  /*@out@*/ struct Chunk *key, /*@out@*/ struct Chunk *val,
	  /*@nonnull@*/ const struct BTree_SeqOption *option)
{
    DB *db = tree.impl->db;
    DBT k, v;
    int s;

    if ((s = db->seq(db, &k, &v, option->flags)) != 0)
	return s;
    Chunk_setDBT(key, k);
    Chunk_setDBT(val, v);
    return 0;
}

static const struct BTree_SeqOption FIRST = {R_FIRST};
static const struct BTree_SeqOption LAST = {R_LAST};
static const struct BTree_SeqOption NEXT = {R_NEXT};
static const struct BTree_SeqOption PREV = {R_PREV};
const struct BTree_SeqOption *const BTree_FIRST = &FIRST;
const struct BTree_SeqOption *const BTree_LAST = &LAST;
const struct BTree_SeqOption *const BTree_NEXT = &NEXT;
const struct BTree_SeqOption *const BTree_PREV = &PREV;
